/*
 * g-page Copyright (C) 2001 Hiro Protagonist
 *
 * This software is released under the GNU General Public License.
 * Please read the included LICENSE file for more information.
 * This software comes with no warranty of any kind, use at you own risk!
 */        

#include "constants.h"
#include "callbacks.h"

static gint load_file(FILE *fp, gint fd);
static void release_mutex();

/*********************************************************************************************/
/**                                                                                         **/
/*    FUNCTION: get_resource_data takes 4 arguments:                                         *
 *                                                                                           *
 *    file_name: The name of the resource file                                               *
 *    match: The resource option we are looking for                                          *
 *    return_string: The string that contains the option we are looking for                  *
 *    data: this is where the resource option gets copied                                    *
 *                                                                                           *
 *    NOTES:                                                                                 *
 *          We try to match a string that contains the option we are looking for. The        *
 *          function assumes the resource file stores the data in tuples: OPTION=data        *
 *                                                                                           * 
 *    RETURN VALUES:                                                                         *
 *    0 if we found a match                                                                  *
 *                                                                                           * 
 *    1 if we can't match the option. The matching string gets copied into the string        *               
 *    (return_string) for the calling function to take action on.                            *
 *                                                                                           * 
 *    2 if we match the string but are unable to parse the data from the tuple               *
 *                                                                                           */
/*********************************************************************************************/

gchar *get_resource_data(gchar *filename, gchar *match, gint max_size)
{

    FILE   *fp;
    gchar  *temp_buffer;
    gchar  *string_array[NUM_FIELDS];
    gchar  *data = NULL;
    gint   index;


    if ( (fp = get_file_pointer(filename, "r")) == NULL)
	return(NULL);

    for (index = 0; index < 2; index++)
	string_array[index] = g_malloc(MAX_LENGTH);
    
    while( (temp_buffer = read_file_by_line(fp)) != NULL)
    {
	if ( (match_in_string(temp_buffer, match, 2, 1, '='))	== 0)
	{
	    fclose(fp);
            if ( (get_fields_from_string(string_array, temp_buffer, 2, '=')) == 1) 
	    {
		g_free(string_array[0]);
		g_free(string_array[1]);
		g_free(temp_buffer);
		return (NULL);
	    }
	    if ( strlen(string_array[1]) >= max_size )
	    {
		g_free(string_array[0]);
		g_free(string_array[1]);
		g_free(temp_buffer);
		return (NULL);
	    }
	    else
	    {
		data = g_strdup(string_array[1]);
		g_free(string_array[0]);
		g_free(string_array[1]);
		g_free(temp_buffer);
		return (data);
	    }
	}
	g_free(temp_buffer);
    }

    fclose(fp);
    g_free(string_array[0]);
    g_free(string_array[1]);

    return (NULL);			/* No match */
	
}


/*********************************************************************************************/
/**                                                                                         **/
/*    FUNCTION: read_file_by_line takes 2 arguments:                                         *
 *                                                                                           *
 *    fp: The file pointer to the stream we are reading from                                 *
 *    return_string: The string holds line of data we just read                              *
 *                                                                                           *
 *    NOTES:                                                                                 *
 *           We are adding a memset to an fgets to make sure we have no garbage in the       *
 *           string.                                                                         *
 *                                                                                           *
 *    RETURN VALUES:                                                                         *
 *     The string if we are successful, NULL on end of file                                              *
 *                                                                                           */
/*********************************************************************************************/

gchar  *read_file_by_line(FILE *fp)
{
    gchar   *return_string;

    return_string = g_malloc(MAX_LENGTH);

    if ( (fgets(return_string, MAX_LENGTH, fp)) == NULL)
    {
	g_free(return_string);
	return (NULL);
    }
    else
	return (return_string);

}

/*********************************************************************************************/
/**                                                                                         **/
/*    FUNCTION: get_fields_from_string takes 4 arguments                                     *
 *                                                                                           *
 *    return_string: An array of pointers to the fields we parsed                            *
 *    string_to_parse: The string we are going to parse                                      *
 *    num_fields: The number of fields in the string                                         *
 *    delimiter: The character that is used to separate the fields                           *
 *                                                                                           *
 *    NOTES:                                                                                 *
 *          We parse the given string and load each parsed field into an array of strings    *
 *          The newline character delimits the last field.                                   *
 *                                                                                           * 
 *    RETURN VALUES:                                                                         *
 *    0 on success                                                                           *
 *                                                                                           * 
 *    1 if we have an error (a field has exceeded the maximum string length or we didn't     *
 *    load the requested number of fields - bad delimiter - corrput database                 *
 *                                                                                           */
/*********************************************************************************************/
    
int  get_fields_from_string(char *return_string[NUM_FIELDS], char string_to_parse[MAX_LENGTH], int num_fields, char delimiter)
{
    gint   string_index, field_index, i;
    gint   records_copied = 0;

    string_index = field_index = i = 0;

    while( string_to_parse[string_index] )
    {
	if( (string_to_parse[string_index] == delimiter) || (string_to_parse[string_index] == '\n') )
	{
	    *(return_string[field_index] + i) = '\0';
	    field_index++;
	    
	    i = 0;
	    string_index++;
	    records_copied++;
	    if ( records_copied == num_fields )
		break;
	}
	else
	    if (i == MAX_LENGTH)
		return 1;	/* Corrput database - Illegal field length */

	*(return_string[field_index] + i++) = string_to_parse[string_index++];
    }
    if(records_copied != num_fields) /* We have a corrupt database */
	return 1;
    return 0;
}

/*********************************************************************************************/
/**                                                                                         **/
/*    FUNCTION: edit_item takes 6 arguments:                                                 *
 *                                                                                           *
 *    item_id: The id number of the string we're editing                                     *
 *    file_name: The name of the file we're editing                                          *
 *    number_of_fields: The number of fields in the string                                   *
 *    item_field: The field number we are looking to match the item_id in                    *
 *    new_string: The string we are going to use to replace the string being edited          *
 *    action_flag: This is the flag to tell the function if we are editing a string or       *
 *                 deleting a string.                                                        *
 *                                                                                           *
 *    NOTES:                                                                                 *
 *          edit_item can delete the string it matches or replace it with new_string.        *
 *          If the action_flag is EDIT then we edit, otherwise we assume delete.             *
 *                                                                                           * 
 *    RETURN VALUES:                                                                         *
 *    0 if we edited or deleted,  1 if we took no action (This indicates an error)           *
 *                                                                                           */
/*********************************************************************************************/
    
int edit_item(char *item_id, char *file_name, int number_of_fields, int item_field, char *new_string, int action_flag, char delimiter)
{

    FILE    *read_fp;
    FILE    *temp_fp;
    gint    return_value;
    gint    good_data = 0;
    gint    file_read;
    gchar   *temp;
    gchar   *temp_line;
    
    if ( (read_fp = get_file_pointer(file_name, "r")) == NULL)
	return(1);

    temp_fp = get_temp_file();
    if ( temp_fp == NULL )
	return (1);

    temp = g_malloc(MAX_LENGTH_GROUP_DATA);

    (char *)file_read = fgets(temp, MAX_LENGTH_GROUP_DATA, read_fp);
    while ( (int *)file_read != NULL )
    {
	return_value = match_in_string(temp, item_id, number_of_fields, item_field, delimiter);

	if (return_value == 1)	/* We didn't match the field  */
	{
	    fputs(temp, temp_fp);
	}
	else			/* We found a match - take action */
	{
	    if (action_flag == EDIT)
		fputs(new_string, temp_fp);
	    good_data++;
	}

	(char *)file_read = fgets(temp, MAX_LENGTH_GROUP_DATA, read_fp);
    }

/*
 *  This code should let me add new items to the config file without 
 *  having to deal with any backward compatability issues.
 */
    
    if ( (good_data == 0) && (action_flag == EDIT) )
    {
	fputs(new_string, temp_fp);
	good_data++;
    }

    fclose(read_fp);

    temp_line = g_malloc(MAX_LENGTH);
    rewind(temp_fp);

    if ( (read_fp = get_file_pointer(file_name, "w")) == NULL)
    {
	g_free(temp_line);
	return 1;
    }
    while (fgets(temp_line, MAX_LENGTH, temp_fp))
	fputs(temp_line,read_fp);
    g_free(temp_line);
    fclose(read_fp);
    fclose(temp_fp);

    g_free(temp);

    if (good_data == 1)		/* success */
	return(0);
    else			/* No match - no action was taken */
	return(1);
}


/*********************************************************************************************/
/**                                                                                         **/
/*    FUNCTION: match_in_string takes 5 arguments:                                           *
 *                                                                                           *
 *    buffer:  The string we are going to search in                                          *
 *    match:   The string we are going to search for                                         *
 *    fileds:  The number of comma delimited fields in the string (buffer)                   *
 *    match_field:  The field we are trying to match                                         *
 *    delimiter:    The character used to delimit each field                                 *
 *                                                                                           *
 *                                                                                           *
 *    NOTES:                                                                                 *
 *         We are trying to match a specified field with a string (match) which gets passed  *
 *         from the calling function. We know what field to look in (match_field)            *
 *                                                                                           *
 *    RETURN VALUES:                                                                         *
 *     0 if we are successful, 1 on no match                                                 */
/*********************************************************************************************/

int match_in_string(char buffer[MAX_LENGTH], char *match, int fields, int match_field, char delimiter)
{
    gint    string_length;
    gint    x, y;
    gint    fields_counted = 0;
    gchar   *temp;
    
    string_length = strlen(buffer);
    temp = g_malloc(string_length + 1);

    x = 0;
    y = 0;

    while(1)
    {
	if ( x > string_length )     /* Error: Something evil has happened */
	{
	    g_free(temp);
	    return 1;		
	}

	if ( (buffer[x] == delimiter) || (buffer[x] == '\n') )  /* End of a field,  */
	{
	    if ( ++fields_counted == match_field )        /* Is this the field we're looking for? */
	    {
		temp[y] = '\0';
		if ( !strcmp(match, temp) )
		{
		    g_free(temp);
		    return 0;
		}
		else
		{
		    g_free(temp);
		    return 1;
		}
	    }
	    else		          /* Not the correct field, try the next one */
	    {
		if ( buffer[x] == '\n' )  /* This should never happen, better safe than sorry */
		    return 1;
		else
		{
		    x++;
		    y = 0;
		}
	    }
	}
	else
	    temp[y++] = buffer[x++];
    }

}

/*********************************************************************************************/
/**                                                                                         **/
/*    FUNCTION: add_data takes 2 arguments:                                                  *
 *                                                                                           *
 *    str:  The string we are adding                                                         *
 *    file_name: The name of the file we are adding the string to                            *
 *                                                                                           *
 *    NOTES:                                                                                 *
 *           We are appending the file then returning                                        *
 *                                                                                           *
 *    RETURN VALUES:                                                                         *
 *     0 if we are successful, 1 on error                                                    *
 *                                                                                           */
/*********************************************************************************************/

int add_data(char *str, char *file_name)
{

    FILE    *fp;
    gchar   *temp;

    if ( (fp = get_file_pointer(file_name, "a")) == NULL)
	return(1);

    temp = g_strdup_printf("%s\n", str);
    fputs(temp, fp);

    fclose(fp);
    free(temp);
    
    return 0;
}

gint delete_receivers_with_service_id(char *service_id)
{
    FILE  *read_fp, *temp_fp;
    gchar *buffer;
    gchar user_id[MAX_FIELD_LENGTH];
    gint  index = 0;
    gint  deleted_receiver = 0;
    gchar *t_file;
    
    memset(user_id, '\0', sizeof(user_id));

    if ( (read_fp = get_file_pointer("RECEIVER.dat", "r")) == NULL)
	return 0;

    temp_fp = get_temp_file();
    if ( temp_fp == NULL )
	return 0;

    while ( (buffer = read_file_by_line(read_fp)) != NULL)
    {
	if ( (match_in_string(buffer, service_id, RECEIVER_FIELDS, 4, ',') ))
	{
	    fputs(buffer, temp_fp);
	    g_free(buffer);
	}
	else
	{
	    while( buffer[index] != ',' )
	    {
		user_id[index] = buffer[index];
		index++;
	    }
	    user_id[index] = '\0';
	    delete_receiver_from_groups(user_id);
	    recycle_id("RECEIVER", user_id);
	    index = 0;
	    memset(user_id, '\0', sizeof(user_id));
	    g_free(buffer);
	    deleted_receiver++;
	}
    }
    fclose(read_fp);
    rewind(temp_fp);


    t_file = g_malloc(MAX_LENGTH);
    if ( ( read_fp = get_file_pointer("RECEIVER.dat", "w")) == NULL)
    {
	fclose(temp_fp);
	g_free(t_file);
	return 0;
    }
    while (fgets(t_file, MAX_LENGTH, temp_fp))
      fputs(t_file, read_fp);

    fclose(read_fp);
    fclose(temp_fp);
    g_free(t_file);
    
    return (deleted_receiver);
}

char *get_field(char buffer[MAX_FIELD_LENGTH], int number_of_fields, int field)
{

    gint  field_counter;
    gint  index = 0;
    gint  counter = 0;
    gchar return_data[MAX_FIELD_LENGTH];
    gchar *data;

    memset(return_data, '\0', sizeof(return_data));
    field_counter = 0;
    while ( (field_counter < (field - 1)) )
    {
	if(buffer[index++] == ',')
	    field_counter++;
    }

    while ( (buffer[index] != ',') && (buffer[index] != '\n') && (buffer[index] != '\0' ) )
	return_data[counter++] = buffer[index++];
/*
    data = g_malloc(MAX_FIELD_LENGTH);
    strcpy(data, return_data);
*/
    data = g_strdup(return_data);
    return data;
}

gchar *get_uid_line(gchar *file_name, gchar *uid, gint num_fields, gint group_flag, gint mem_size)
{

    FILE  *f_pointer;
    char  *blah = "";
    gchar  *(* read_file)();
    gchar  *line;

    if ( (f_pointer = get_file_pointer(file_name, "r")) == NULL)
	return(NULL);

    if (group_flag)
	read_file = read_big_file_by_line;
    else
	read_file = read_file_by_line;

    while ( (line = ( (* read_file)(f_pointer, mem_size))) != NULL)
    {
	blah = get_field(line, num_fields,0);
	if  ( !(strcmp(uid, blah)) )
	{
	    fclose(f_pointer);
	    g_free(blah);
	    return (line);
	}
	else
	{
	    g_free(blah);
	    g_free(line);
	}
    }
    
    fclose(f_pointer);
    fprintf(stderr, "Error: Corrupt database %s\n", file_name);
    return(NULL);
}
    
gchar *read_big_file_by_line(FILE *fp, gint mem_size)
{
    gint  index, mem_left;
    gint  c;
    gchar *return_string;

    enum { CHUNK_SIZE = 500 };

    index = 0;
    mem_left = mem_size;
    return_string = g_malloc(mem_size);

    while ( (c = fgetc(fp)) != '\n' )
    {
	if ( c == EOF )
	{
	    g_free(return_string);
	    return( return_string = NULL );
	}
	if ( mem_left == 0 )
	{
	    if ( !(return_string = g_realloc(return_string, mem_size + CHUNK_SIZE)) ) /* no need to save the results it we get a null */
	    {
		g_free(return_string);
		return (NULL); 
	    }
	    else
	    {
		mem_left = CHUNK_SIZE;
		mem_size += CHUNK_SIZE;

		*(return_string + index++) = c;
		mem_left--;
	    }
	}
	else
	{
	    *(return_string + index++) = c;
	    mem_left--;
	}
    }

/*
 *    We need to make sure we have room for the newline and null terminator
 */

    if ( mem_left < 2 )
    {
	if ( !(return_string = g_realloc(return_string, mem_size + 2)) ) 
	{
	    g_free(return_string);
	    return (NULL); 
	}
	else
	{
	    *(return_string + index++) = '\n';
	    *(return_string + index++) = '\0';
	    
	    fprintf(stderr, "End mem size: %d\n", mem_size + 2);
	}
    }
    else
    {
	*(return_string + index++) = '\n';
	*(return_string + index) = '\0';
    }


    return(return_string);
}
	
FILE *get_file_pointer(char *file_name, char *options)
{

    FILE         *fp;
    extern  gint network_db;

    gchar        *error_msg;
    gchar        *current;
    gchar        *home;
    gchar        *dir_path;
    gchar        *filename;

    filename = g_strdup(file_name);

    if ( network_db )
    {
	if ( !strcmp(filename, "RECEIVER.dat") )
	    strcpy(filename, "RECEIVER.net");
	else if ( !strcmp(filename, "SERVICE.dat") )
	    strcpy(filename, "SERVICE.net");
	else if ( !strcmp(filename, "GROUP.dat") )
	    strcpy(filename, "GROUP.net");
	else if ( !strcmp(filename, "IDS.dat") )
	    strcpy(filename, "IDS.net");
    }

    error_msg = g_strdup_printf("Error opening %s", filename);

    home = g_strdup(getenv("HOME"));

    current = get_working_dir();

/*
 *   CONFIG.dat has to be located in ~/.gpage, thus we must check to see if the user
 *   is trying to open it or another file, which might be located anywhere
 */

    if ( !strcmp(filename, "CONFIG.dat") )
	dir_path = g_strdup_printf("%s/.gpage/%s", home, filename);
    else
	dir_path = g_strdup_printf("%s%s", current, filename);

    if ( ( fp = fopen(dir_path, options)) == NULL)
    {
	if ( network_db )
	    generic_error(" Error opening network database files ", "G-Page Error");
	fprintf(stderr, "(get file pointer) Error opening %s\n", filename);
	//update_status_line(error_msg);
    }

    g_free(current);
    g_free(home);
    g_free(error_msg);
    g_free(dir_path);
    g_free(filename);

    return (fp);
}

FILE   *get_temp_file()
{
    FILE    *fp;

    fp = tmpfile();    
    return (fp);
}

gchar *get_working_dir()
{
    FILE   *fp;
    gchar  *cdir;
    gchar  *current;
    gchar  *resource_string;
    gchar  *temp_buffer;
    gint   index;
    gchar  *string_array[NUM_FIELDS];
    
    current = g_malloc(MAX_LENGTH);
    resource_string = g_malloc(MAX_LENGTH);

    cdir = g_strdup_printf("%s/.gpage/CONFIG.dat", getenv("HOME") );

    if ( (fp = fopen(cdir, "r")) == NULL)
    {
	perror("(cdir) Error opening CONFIG.dat");
	exit(EXIT_FAILURE);
    }

    for (index = 0; index < 2; index++)
	string_array[index] = g_malloc(MAX_LENGTH);
    
    while( (temp_buffer = read_file_by_line(fp)) != NULL)
    {
	if ( (match_in_string(temp_buffer, "db_dir", 2, 1, '='))== 0)
	{
            if ( (get_fields_from_string(string_array, temp_buffer, 2, '=')) == 1) 
	    {
		g_free(string_array[0]);
		g_free(string_array[1]);
		g_free(cdir);
		g_free(resource_string);

		fclose(fp);
		update_status_line("Error reading value for db_dir in CONFIG.dat");
		g_free(temp_buffer);
		return(NULL);
		
	    }
	    strcpy(current, string_array[1]);

	    g_free(string_array[0]);
	    g_free(string_array[1]);
	    g_free(cdir);
	    g_free(resource_string);
	    g_free(temp_buffer);
	    
	    fclose(fp);    
	    return (current);

	}
	g_free(temp_buffer);
    }

    sys_error("Error corrupt file: CONFIG.dat", EXIT);
    return(NULL);		/* We should never get here */
}

void sys_error(gchar *error_msg, gint action)
{
    switch (action)
    {
    case UPDATE:
	fprintf(stderr, "%s\n", error_msg);
	update_status_line(error_msg);
	break;

    case PERROR:
	perror(error_msg);
	exit(EXIT_FAILURE);
	break;

    case EXIT:
	fprintf(stderr, "%s\n", error_msg);
	exit(EXIT_FAILURE);
	break;
    }    
}

/*
 *   We return a one if we are unable to load the network database files, 
 *   we return a zero otherwise.
 */

int get_network_database()
{ 

    enum {
        RECEIVER = 1,
        SERVICE = 2,
        GROUP = 3,
        IDS = 4
    };

    extern pthread_mutex_t  update_live_page_mutex;

    extern gint             network_db;
    extern gint             live_page;

    FILE                    *fp;

    struct  sockaddr_in     address;
    struct  in_addr         inaddr;
    struct  hostent         *host;

    gint                    not_done, file;             
    gint                    sock;
    gint                    mutex_status;
    gint                    num_read;
    gchar                   *port_data, *domain, *permission;
    gchar                   filename[512];

/*
 *     We need the mutex here so we don't try to update the database
 *     while a page is being sent. This could lead to some undefined 
 *     behaviour if the pager info of the pager we are sending to
 *     is changed or deleted
 */

    if ( (mutex_status = pthread_mutex_lock(&update_live_page_mutex)) != 0 )
    {
	perror("Error locking mutex - possible data corruption");
	return (1);
    }

    if ( live_page )   /* This should never happen - we shouldn't get a mutex if this is over zero */
    {
	if ( (pthread_mutex_unlock(&update_live_page_mutex)) != 0 )
	{
	    perror("Error unlocking mutex - possible data corruption");
	    update_status_line("Error: unable to free live page mutex - please restart g-page");
	    generic_error("\n An unrecoverable error has occured \n Please close g-page and restart to continue \n", "G-Page Error");
	}

	return (1);
    }


/*
 *   After this point we assume that an error means we cannot load the network database, 
 *   if an error occurs we will set the global network_db to zero and reload the local database
 */

    if ( (domain = get_resource_data("CONFIG.dat", "database_adr", MAX_FIELD_LENGTH)) == NULL)
    {
	update_status_line("Error: corrupt file: CONFIG.dat");
	network_db = 0;
	release_mutex();
	return (1);
    }


    if ( (port_data = get_resource_data("CONFIG.dat", "database_port", MAX_FIELD_LENGTH)) == NULL)
    {
	update_status_line("Error: unable to read port from CONFIG.dat");
	network_db = 0;
	release_mutex();

	return (1);
    }


    if (inet_aton(domain, &inaddr))
	host = gethostbyaddr((char *) &inaddr, sizeof(inaddr), AF_INET);
    else
	host = gethostbyname(domain);    
    g_free(domain);

    if (!host)
    {
	generic_error("\n  Unable to lookup host (bad address?) \n", "G-Page Error");
	network_db = 0;
	release_mutex();

	return(1);
    }

    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
    {
	update_status_line("Error creating socket ");
	network_db = 0;
	release_mutex();

	return(1);
    }

    address.sin_family = AF_INET;
    
    address.sin_port = htons(atoi(port_data));
    g_free(port_data);

    memcpy(&address.sin_addr, host->h_addr_list[0], sizeof(address.sin_addr));

/*
 *      connect_no_block takes the same arguments as connect except we add a timeout value at the end
 *      This allows us to set a smaller value than the default one. 
 */

    if ( connect_no_block(sock, (struct sockaddr *) &address, sizeof(address), 30) )
    {				/* No Connect  */
	if ( errno == ETIMEDOUT )
	    generic_error("\n Error connecting with database server \n Connection timed out \n", "G-Page Error");
	else
	    generic_error("\n  Unable to connect to remote workstation \n", "G-Page Error");
	    
	network_db = 0;
	release_mutex();
	return(1);
    }
    else
    {
	permission = g_malloc(MAX_LENGTH);

	/*  The first thing the db server sends us is our permission status - denied if we are banned, okay otherwise */

	num_read = read_a_line(sock, permission, MAX_LENGTH);
	if ( num_read == 0 )
	{
	    generic_error("\n Error reading from database server \n", "G-Page Error");
	    close(sock);
	    g_free(permission);
	    network_db = 0;
	    release_mutex();
	    
	    return(1);
	}


	if ( !strcmp(permission, "denied\n") )
	{
	    generic_error("\n  Permission to connect with database server: DENIED  \n", "G-Page Error");
	    close(sock);
	    g_free(permission);
	    network_db = 0;
	    release_mutex();

	    return(1);
	}
	g_free(permission);
	
	not_done = 1;
	file = RECEIVER;

	while ( not_done )
	{
	    switch (file) {
	    case RECEIVER:
		strcpy(filename, "RECEIVER.net");
		break;

	    case SERVICE:
		strcpy(filename, "SERVICE.net");
		break;

	    case GROUP:
		strcpy(filename, "GROUP.net");
		break;

	    case IDS:
		strcpy(filename, "IDS.net");
		not_done = 0;

		break;
	    default:
		not_done = 0;
		break;
            }

	    
	    if ( (fp = get_file_pointer(filename, "w")) == NULL )
	    {
		network_db = 0;
		not_done = 0;
	    }
	    else
	    {
		if ( load_file(fp, sock) )
		{
		    generic_error("\n  Error reading database files \n  Connection terminated by server \n", "G-Page Error");
		    not_done = 0;
		    network_db = 0;
		}
		else
		    file++;
	    }

	} /* End of the switch block */
	close(sock);
    } /* End of the while loop */

    release_mutex();
    
    update_status_line_small("Selected receivers: 0");	/* When we update the db we lose all receiver selections  */

    if ( network_db )		/* Check to see if an error  occured while reading the files */
	return (0);
    else
	return (1);
    
}

static gint load_file(FILE *fp, gint fd)
{
    enum {
	READ,
	WRITE
    };

    gchar   *buffer;
    gint    not_done;
    gint    num;
    gint    maxfdp1;
    gint    result;
    gint    num_read;
    gint    state;
    fd_set  rset, wset;
    char    *end = "**********\n"; 

    num_read = result = 0;

    maxfdp1 = fd + 1;

    not_done = 1;
    state = READ;

    buffer = g_malloc(MAX_LENGTH);
    memset(buffer, '\0', MAX_LENGTH);

    while (not_done)
    {

/*
 *   First we read a line from the database server
 *   We need to use select here because the db server simply closes it's connection
 *   if an un recoverable error occurs
 */

	switch (state)
	{
	case WRITE:
	    FD_ZERO(&rset);
	    FD_SET(fd, &rset);
	    result = select(maxfdp1, &rset, NULL, NULL, NULL);
	    if ( result == -1 )
	    {
		if ( errno == EINTR )
		{
		    errno = -1;
		    break;
		}
		fclose(fp);
		g_free(buffer);
		return (1);
	    }
	    else
	    {
		if ( FD_ISSET(fd, &rset) )
		{
		    num_read = read_a_line(fd, buffer, MAX_LENGTH);
		    if ( num_read == 0 ) /* we have a closed socket */
		    {
			fclose(fp);
			g_free(buffer);
			return(1);
		    }
		}
		else		/* should not happen */
		{
		    fclose(fp);
		    g_free(buffer);
		    return(1);
		}
	    }
	    state = READ;
	    break;
	    
/*
 *   Now we write the line back to the db server, then if we save it to file locally
 */

	case READ:
	    FD_ZERO(&wset);
	    FD_SET(fd, &wset);
	    result = select(maxfdp1, NULL, &wset, NULL, NULL);
	    if ( result == -1 )
	    {
		if ( errno == EINTR )
		{
		    errno = -1;
		    break;
		}

		g_free(buffer);
		fclose(fp);
		return (1);
	    }
	    else
	    {
		if ( FD_ISSET(fd, &wset) )
		{
		    num = write_num(fd, buffer, strlen(buffer));
		    if ( !strcmp(buffer, end) )
			not_done = 0;
		    else
			fputs(buffer, fp);
		}
		else
		{
		    g_free(buffer);
		    fclose(fp);
		    return (1);
		}
	    }
	    state = WRITE;
	    break;

	}
    }
    fclose(fp);
    g_free(buffer);    
    return (0);
}

gint network_db_refresh()
{
    extern    gint  network_db;

    if ( network_db )
    {
	get_network_database();
	return TRUE;
    }
    else
	return FALSE;
}


int connect_no_block(int sockfd, struct sockaddr *serv_addr, int addrlen, int timeout)
{

    gint             flags, n, error;
    socklen_t        len;
    fd_set           rset, wset;
    struct timeval   tval;

    if ( (flags = fcntl(sockfd, F_GETFL, 0)) == -1 )
	return (-1);

    if ( (fcntl(sockfd, F_SETFL, flags | O_NONBLOCK)) == -1 )
	return (-1);

    error = 0;

    if ( (n = connect(sockfd, (struct sockaddr *) serv_addr, addrlen)) < 0)
	if ( errno != EINPROGRESS )
	    return (-1);

    if ( n == 0 )		/* connect completed immediatly */
	goto done;

    FD_ZERO(&rset);
    FD_SET(sockfd, &rset);
    wset = rset;
    tval.tv_sec = timeout;
    tval.tv_usec = 0;

    if ( ( n = select(sockfd + 1, &rset, &wset, NULL, timeout ? &tval : NULL)) == 0 )
    {			/* Timeout!! */
	close(sockfd);
	errno = ETIMEDOUT;
	return (-1);
    }
    
    if (FD_ISSET(sockfd, &rset) || FD_ISSET(sockfd, &wset))
    {
	len = sizeof(error);
	if ( getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len) < 0)
	    return (-1);		/* Solaris pending error */
    }
    else
	return (-1);		/* select error */

    
 done:
    if ( (fcntl(sockfd, F_SETFL, flags)) == -1)
	perror("Error resetting fcntl flags: "); /* We'll ignore this and continue on */
	  
    if (error)
    {
	close(sockfd);
	errno = error;
	return (-1);
    }

    return (0);
}

static void release_mutex()
{
    extern pthread_mutex_t  update_live_page_mutex;
    extern GtkWidget        *receiver_clist; 
    extern gint             network_db;

    if ( (pthread_mutex_unlock(&update_live_page_mutex)) != 0 )
    {
	update_status_line("Error: unable to free live page mutex - please restart g-page");
	generic_error("\n An unrecoverable error has occured \n Please close g-page and restart to continue \n", "G-Page Error");
	return;
    }
    
    if ( network_db == 0 )	/* an error occured somewhere - reset database to local */
	edit_item("database", "CONFIG.dat", 2, 1, "database=local\n", EDIT, '=');	  

    gtk_clist_freeze(GTK_CLIST(receiver_clist));
    gtk_clist_clear(GTK_CLIST(receiver_clist));
    load_receiver_list("RECEIVER.dat", GTK_CLIST(receiver_clist));
    load_receiver_list("GROUP.dat", GTK_CLIST(receiver_clist));
    load_main_list_with_pixmaps();
    gtk_clist_thaw(GTK_CLIST(receiver_clist));

}
