/*
 * g-page Copyright (C) 2000 Hiro Protagonist
 *
 * This software is released under the GNU General Public License.
 * Please read the included LICENSE file for more information.
 * This software comes with no warranty of any kind, use at you own risk!
 */        

#include "constants.h"
#include "callbacks.h"

ssize_t read_num(int file_desc, char *strptr, size_t num)
{

  size_t      num_left;
  ssize_t     num_read;
  gchar        *ptr;
  
  ptr = strptr;
  num_left = num;
  while ( num_left > 0 )
    {
      if ( (num_read = read( file_desc, ptr, num_left ) ) < 0 )
	{	
	  if ( errno == EINTR )
	    num_read = 0;
	  else
	    return (-1);
	}
      else
	if ( num_read == 0)
	  break;     /* We have reached the end of the file  */
      num_left -= num_read;
      ptr += num_read;
    }
  return ( num - num_left );
}


ssize_t write_num(int file_desc, char *strptr, size_t num)
{

  size_t             num_left;
  ssize_t            num_written;
  gchar              *ptr, *ptr1;

  ptr = g_malloc(sizeof(*ptr));
  ptr1 = ptr;
  
  ptr = strptr;
  num_left = num;
  while ( num_left > 0 )
  {
      if ( ( num_written = write(file_desc, ptr, num_left )) <= 0 )
      {
	  if ( errno == EINTR )
	    {
	      num_written = 0;
	      continue;
	    }
	  else
	  {
	      g_free(ptr1);
	      return (-1);
	  }
      }
      num_left -= num_written;
      ptr += num_written;
  }
  g_free(ptr1);
  return (num);
}

/*
 *  FUNCTION READ_A_LINE
 *  
 *  This function takes an open file descriptor and reads characters one
 *  at a time from it until it hits a newline "\n". We store the string into
 *  return_buff and return the number of characters read. This is not
 *  efficient but with the small amounts of data we are reading this
 *  should not harm us. 
 *
 */

ssize_t read_a_line(int file_desc, char return_buff[MAX_LENGTH], size_t maxlen)
{

    gint  characters_read;
    gint  read_count;
    gchar ch;

    memset(return_buff, (char) '\0', MAX_LENGTH);
    characters_read = 0;
    while (characters_read < (MAX_LENGTH -1 ) )
    {
      read_count = read(file_desc, &ch, 1);
	if ( read_count == 1 )
	{
	    return_buff[characters_read++] = ch;
	    if ( ch == '\n' )
	    {
		return_buff[characters_read++] = '\0';
		return characters_read;
	    }
	}
	else if ( read_count == 0 )
	{
	    if ( characters_read == 0 )
		return 0;	      
	    else
	    {
		return_buff[characters_read++] = '\n';
		return_buff[characters_read++] = '\0';
		return (characters_read);
	    }
	}
	else
	{
	    if ( errno == EINTR )
		continue;
	    else
		return (-1);	
	}
    }
    return_buff[characters_read -2] = '\n';
    return_buff[characters_read -1] = '\0';
    return characters_read;
}

/*
 *  FUNCTION READ_A_BIG_LINE
 *  
 *  This function takes an open file descriptor and reads characters one
 *  at a time from it until it hits a newline "\n". We store the string into
 *  buffer and return the string. The memory is allocated on the fly for
 *  the return string to allow for strings of almost any size.
 *
 */

gchar *read_a_big_line(gint file_desc)
{

    gchar  *buffer;
    gchar  ch;
    gint   characters_saved;
    gint   read_count;
    gint   mem_left, mem_size;

    enum { CHUNK_SIZE = 500 };

    buffer = g_malloc(MAX_LENGTH); /* we start with a number big enough to work for most strings */
    mem_left = MAX_LENGTH;
    mem_size = MAX_LENGTH;

    characters_saved = 0;

    while ( 1 )
    {
	if ( mem_left == 0 )
	{
	    buffer = g_realloc(buffer, mem_size + CHUNK_SIZE);
	    if ( buffer == NULL ) /* If we run out of memory we don't care about saving the string */
		return(NULL);
	    else
	    {
		mem_size += CHUNK_SIZE;
		mem_left = CHUNK_SIZE;
	    }
	}

	read_count = read(file_desc, &ch, 1);
	if ( read_count == 1 )
	{
	    if ( ch == '\n' )
		break;
	    else
	    {
		buffer[characters_saved++] = ch;
		mem_left--;
	    }
	}
	else if ( read_count == 0 )
	{
	    if ( characters_saved == 0 )
	    {
		g_free(buffer);
		return (NULL);	
	    }      
	    else
		break;		/* The file desc is probably closed, lets return what we have so far */
	}
	else
	{
	    if ( errno == EINTR )
		continue;
	    else
	    {
		g_free(buffer);
		return (NULL);	
	    }
	}
    } /* End of infinite while loop */

    if ( mem_left < 2 )
    {
	buffer = g_realloc(buffer, mem_size + 2 );
	if ( buffer == NULL ) /* If we run out of memory we don't care about saving the string */
	    return(NULL);
    }


    buffer[characters_saved++] = '\n';
    buffer[characters_saved] = '\0';
    return (buffer);

}

/*
 *  FUNCTION READ_CACHED_LINE
 *  
 *  This function will take an open file descriptor and read from it until 
 *  there are no characters left to read. It stores the data it reads
 *  into a linked list of strings terminated with newlines. It then 
 *  copies the oldest string into return_buff and returns a one. If 
 *  there is no data to return we copy nothing and return zero. A 
 *  return of -1 indicates an error. The linked list is static so we
 *  retain all extra data that is read in beyond a single line.
 *
 *  We set a timeout on the read function to make sure we don't block 
 *  for too long. 
 *
 *  Problems: This function calls read_a_line which performs the actual
 *  reading of data at one character at a time. To make this more efficient
 *  this needs to call another function which reads in blocks of MAX_LENGTH
 *  then parses out the strings ending with newlines and loades them 
 *  into the linked list. This is not a major problem since we are dealing
 *  with small amounts of data but at some point in time it needs to be 
 *  written to be more efficient
 *
 */

gint read_cached_line(gint file_desc, gchar *return_buff, gint max_length)
{

    typedef struct cache {
	gchar *data;
	struct cache *next;
    }cache;

    struct timeval    time_out;
    gint              num_read, maxfdp1;
    fd_set            read_set;
    gchar             *read_buff;
    gint              select_result;
    static cache      *temp_node;
    static cache      *head;
    static cache      *tail;
    static gint       first = 1;

    if ( file_desc < 0 )	/* This is an error: closed descriptor */
      return (-1);

    if ( first )
    {
	head = g_malloc(sizeof(*head));
	tail = g_malloc(sizeof(*tail));
	head->next = NULL;
	tail = head;
	first = 0;
    }

    read_buff = g_malloc( MAX_LENGTH + MAX_LENGTH );

    maxfdp1 = file_desc + 1;
    FD_ZERO(&read_set);
    FD_SET(file_desc, &read_set);

    num_read = 0;
    do
    {
	time_out.tv_sec  = 0;
	time_out.tv_usec = 100;	     /* We don't need to block very long here */

	select_result = select(maxfdp1, &read_set, NULL, NULL, &time_out);
	if ( select_result == 0 )   /* Timeout reading from the pipe */
	    num_read = 0;
	else if (select_result < 0 )
	{
	    if (errno == EINTR)
		continue;
	    else
	    {
		//fprintf(stderr, "The bad file descriptor is %d\n", file_desc);
		g_free(read_buff);
		perror("Error in select");
		return (-1);  
	    }
	}
	if (FD_ISSET(file_desc, &read_set))
	{
	    //fprintf(stderr, "The good file descriptor is %d\n", file_desc);
	    num_read = read_a_line(file_desc, read_buff, sizeof(read_buff)); 
	    if ( num_read > 0 )
	    {
		read_buff[(strlen(read_buff) -1)] = '\0'; /* Kill the newline char */
		tail->data = g_malloc(strlen(read_buff) + 1);
		strcpy(tail->data, read_buff);
		/* fprintf(stderr, "READ CACHE %s\n", tail->data); */
		temp_node = g_malloc(sizeof(*temp_node));
		temp_node -> next = NULL;
		tail->next = temp_node;
		tail = temp_node;
	    }
	}
    }while (num_read != 0);

    if ( head != tail )
    {
	strcpy(return_buff, head->data);
	/* fprintf(stderr, "RETURN BUFF: %s\n", return_buff); */
	temp_node = head;
	head = head->next;
	g_free(temp_node->data);
	g_free(temp_node);
	g_free(read_buff);
	return (1);		/* Good data */
    }
    else			/* Head and tail are the same so we have nothing to return */
    {
	strcpy(return_buff, "");
	g_free(read_buff);
	return (0);		/* No data */
    }

}

 





