/*
 * g-page Copyright (C) 2001 Hiro Protagonist
 *
 * This software is released under the GNU General Public License.
 * Please read the included LICENSE file for more information.
 * This software comes with no warranty of any kind, use at you own risk!
 */        

#include "constants.h"
#include "callbacks.h"

gchar *get_page_code(gchar *line);


#define NHASH 500

typedef struct hash_list hash_list;

struct hash_list {
    gchar *name;
    gint  count;
    hash_list *next;
};
hash_list *log_hash[NHASH];

static gint table_insert_lookup(gint hash_value, gchar *name, gint lookup);
static void clean_up_list(void); 
static void parse_log_line(gchar *line, gchar *status, gchar *via, gchar *date, gchar *message, gchar *name, gchar *code);

/*
 *
 *  FUNCTION LOAD_LOG_FOR_LIST
 *
 *  This function reads through the log file and keeps a running
 *  total of the number of times a page was sent to a certain pager
 *  or the number of pages sent on a certain date. The options
 *  variable tells the function to tally by date or by name. The
 *  totals are kept in a linked list which has two elements, a
 *  string which is the name or date and the number of total pages
 *  sent.
 *
 */


void load_log_for_list(log_window_data *log_data, gint options)
{
    gchar  *log_line[3];
    gchar  *name;
    FILE   *fp;
    gchar  *line;
    gint   hash_lookup;
    gint   x;
    list   *log, *head, *temp;

    char  *(* get_arg)();
    
    log = temp = NULL;

    log = get_list_mem(log);

    head = log;
    
    for ( x = 0; x <= NHASH; log_hash[x++] = NULL );

    gtk_clist_freeze(GTK_CLIST(log_data->log_list_clist) );

    gtk_clist_clear( GTK_CLIST(log_data->log_list_clist) );

/*
 *    A function pointer is used here to set the proper fucntion outside of the
 *    while loop, thus making it a bit faster. 
 */

    if ( options == NAME )
	get_arg = get_name;
    else
	get_arg = get_date;

    
    if ( (fp = get_file_pointer("log.gpage", "r")) == NULL)
	return;

    while ( (line = (read_file_by_line(fp))) != NULL)
    {

/*
 *   get_arg is set to either get_name() or get_date() 
 */

	name = (* get_arg)(line);
	hash_lookup = table_insert_lookup(0, name, 0);

/*
 *   We try to insert the string into the hash table. If it did not exist we return a zero
 *   and we need to add it to the main list, otherwise we return the count
 */

	if ( hash_lookup != -1 )
	{
	    temp = get_list_mem(temp);
	    log->next = temp;
	    strcpy(log->name, name);
	    log->hash = hash_lookup;
	    temp->next = NULL;
	    log = log->next;
	}	    
	g_free(name);
	g_free(line);
    }
    
    log = head;
    while(log->next)
    {
	hash_lookup = table_insert_lookup(log->hash, log->name, 1);
	log->count = hash_lookup;
	log = log->next;
    }

    log = head;

/*
 *
 *  We are done counting the number of pages sent so we need
 *  to load up the clist with the name (name can be a pager name
 *  or a date) and the number sent to that name
 *
 */

    log = head;
    while ( log->next != NULL )
    {
	log_line[0] = g_strdup(log->name);

	log_line[1] = g_strdup_printf("%d", log->count);


    	if ( options == NAME )
	    log_line[2] = g_strdup(log->name);
	else
	    log_line[2] = g_strdup("0" );

	gtk_clist_append( (GtkCList *)log_data->log_list_clist, log_line);
	
	g_free(log_line[0]);
	g_free(log_line[1]);
	g_free(log_line[2]);

	temp = log;
	log = log->next;	    
	g_free(temp->name);
	g_free(temp);
    }

    g_free(log->name);
    g_free(log);

    gtk_clist_thaw(GTK_CLIST(log_data->log_list_clist) );
    fclose(fp);  

    clean_up_list();		/* Free the memory in the linked list from the hash table */
}

static void clean_up_list()
{
    extern hash_list *log_hash[NHASH];
    
    hash_list  *temp;
    gint   x;

    for ( x = 0; x < NHASH; x++ )
    {
	if ( log_hash[x] == NULL )
	    continue;
	else
	{
	    while( log_hash[x] )
	    {
		temp = log_hash[x]->next;
		g_free(log_hash[x]->name);
		g_free(log_hash[x]);
		log_hash[x] = temp;
	    }
	}
    }

}

static gint table_insert_lookup(gint hash_value, gchar *name, gint lookup)
{
    extern hash_list *log_hash[NHASH];

    enum {
	MULTIPLIER = 31
    };

    unsigned int   h;
    unsigned char  *p;
    hash_list *hl;

    h = 0;

    for (p = ( unsigned char *)name; *p != '\0'; p++)
	h = MULTIPLIER * h + *p;

    hash_value = h % NHASH;

    for ( hl = log_hash[hash_value]; hl != NULL; hl = hl->next)
    {
	if ( !strcmp(name, hl->name) )
	{
	    if (lookup)
	    {
		return hl->count;
	    }
	    else
	    {
		hl->count++;
		return (-1);
	    }
	}
    }
    
/*
 *   If we get here the name was not in the list we add it and return the hash value
 */

    hl = g_malloc(sizeof(*hl));
    hl->name = g_strdup(name);
    hl->count = 1;
    hl->next = log_hash[hash_value];
    log_hash[hash_value] = hl;

    return (hash_value);
}


list *get_list_mem(list *node)
{
    node = g_malloc(sizeof(*node));
    node->name = g_malloc(MAX_FIELD_LENGTH);
    node->next = NULL;
    
    return node;
}

gchar *get_name(char *log_line)
{
    gint  i;
    gint  x;
    gint  y;
    gchar *name;

    name = g_malloc(MAX_FIELD_LENGTH);

    y = 0;

/*
 *   The name is in the third comma delimited field so we need to skip over the firt two 
 */

    for ( i = 0; i < 3; i++ )
	while( *(log_line + y++) != ',');

/* skip the empty space */

    y++;		

    x = 0;
    while ( *(log_line + y) != ',' )
	*(name + x++) = *(log_line + y++);

    *(name + x) = '\0';
        
    return name;
}

/*
 *
 *  This function just returns the month and the date, it
 *  ignores all other aspects from the date like the year
 *  and the time sent.
 *
 */

char *get_date(char *log_line)
{

    char *short_date = "";
    char one[128], two[128], weekday[128], day[128], five[128], year[128];

    sscanf(log_line, "%s %s %s %s %s %4[^,] ", one, two, weekday, day, five, year);

    short_date = g_strdup_printf("%s %2s  %s", weekday, day, year);

    return short_date;

}

gchar *get_message(gchar *line)
{
    gint  i;
    gint  x;
    gint  y;
    gchar *msg;

    msg = g_malloc(MAX_LENGTH);

    y = 0;

/*
 *   The message is in the fourth comma delimited field so we need to skip over the firt three 
 */

    for ( i = 0; i < 4; i++ )
	while( *(line + y++) != ',');

    y++;		/* skip the empty space */

    x = 0;

    while ( *(line +y) != '\n' )
	*(msg + x++) = *(line +y++);

    *(msg + x) = '\0';

    return msg;
}

/*
 *
 *  FUNCTION LOAD_LOG_LIST
 *
 *  This function reads through the log file and loads all the pertinant
 *  information into a clist for a particular person or a particular date.
 *  The options variable tells the function to use the date or name as a 
 *  match field. 
 *
 */

void load_log_list(GtkWidget *log_clist, gchar *field, gint options)
{
    gchar  *log_line[6];
    gchar  *name, *via, *status, *message;
    gchar  *date = "";
    FILE   *fp;
    gchar  *line;
    gchar  *match;
    gchar  *temp_date;
    gchar  *code;
    gchar  *t_name;
    gint   i;


    if ( (fp = get_file_pointer("log.gpage", "r")) == NULL)
	return;

    match = g_malloc(MAX_FIELD_LENGTH);
    line = g_malloc(MAX_LOG_LENGTH);

    status = g_malloc(16);
    via = g_malloc(16);
    date = g_malloc(32);
    message = g_malloc(MAX_LENGTH);
    name = g_malloc(MAX_FIELD_LENGTH);
    code = g_malloc(16);
     
    while ( (fgets(line, MAX_LENGTH, fp)) != NULL )
    {
	if ( options == NAME )
	{
	    t_name = get_name(line);
	    strcpy(match, t_name);
	    g_free(t_name);
	}
	else
	{
	    temp_date = get_date(line);
	    strcpy(match, temp_date);
	    g_free(temp_date);
	}


	if ( !strcmp(match, field) )
	{
	    parse_log_line(line, status, via, date, message, name, code);

	    log_line[0] = g_strdup(status);
	    log_line[1] = g_strdup(via);

	    if ( options == DATE )
		log_line[2] = g_strdup((date + 7));
	    else
		log_line[2] = g_strdup(date);

	    if ( options == NAME )
	    {
		log_line[3] = g_strdup(message);
		log_line[4] = g_strdup("");
	    }
	    else
	    {
		log_line[3] = g_strdup(name);
		log_line[4] = g_strdup(message);
	    }

	    log_line[5] = g_strdup( code);

	    gtk_clist_append( (GtkCList *)log_clist, log_line);
	    
	    for ( i = 0; i < 6; i++ )
		g_free(log_line[i]);
	}
    }

    g_free(status);
    g_free(via);
    g_free(date);
    g_free(code);
    g_free(message);
    g_free(name);

    g_free(line);
    fclose(fp);
    g_free(match);
		    
}

/*
 *  FUNCTION GET_LOG_DATE
 *
 *  This function parses the entire date from the line passed to it.
 *  The date includes the month, date, time and year.
 *
 */

gchar *get_log_date(gchar *line)
{
    gint  i = 0;
    gint  x = 0;
    gchar *ret_date;

    ret_date = g_malloc(32);

    while ( *(line + i++) != ',');
    i++;
    while ( *(line + i++) != ',');
    i++;
    while ( *(line + i++) != ' ');
    
    while ( (*(ret_date + x++) = *(line + i++)) != ',' );
    *(ret_date + --x) = '\0';
    return (ret_date);

}

/*
 *  FUNCTION GET_LOG_STATUS
 *
 *  We are looking at a three character string which is the return code
 *  from a page that was sent. The first char tells us if it was good
 *  page or if it failed. A zero indicates a good page, anything else 
 *  means we failed.
 *
 *  At some point in time this will have to look at the char for the failed
 *  pages to report back why it failed, for now we just report that it failed.
 *
 */

gchar *get_log_status(gchar *line)
{
    gchar   *status;
    gint    i = 0;
    
    status = g_malloc(16);

/*
 *   Skip the epoch string
 */

    while ( line[i++] != ',' );

    if ( line[i] == '0' )
	strcpy(status, "sent");
    else
	strcpy(status, "failed");

    return status;
}

/*
 *  FUNCTION GET_VIA
 *
 *  We are looking at a three character string which is the return code
 *  from a page that was sent. The second char tells us if it was sent 
 *  via snpp or email. A zero indicate email, a one indicates snpp.
 *
 */

gchar *get_via(gchar *line)
{
    gchar  *via;
    gint   i = 0;

    via = g_malloc(16);

/*
 *   Skip the epoch string
 */

    while ( line[i++] != ',');
    i++;

    if ( line[i] == '1' )
	strcpy(via, "email");
    else if ( line[i] == '2' )
	strcpy(via, "snpp");
    else if ( line[i] == '3' )
	strcpy(via, "wctp");

    return via;
}


gchar *get_page_code(gchar *line)
{
    gchar  *code;
    gint   i = 0;
    gint   x = 0;

    code = g_malloc(16);

/*
 *   Skip the epoch string
 */

    while ( line[i++] != ',');

    x = 0;
    while ( line[i] != ',')
	*(code + x++) = line[i++];

    *(code + x) = '\0';
    return code;
}

gint delete_log_data(char *name, gint what)
{
    FILE    *fp, *fpt;
    gint    return_status = 0;
    gchar   *line = NULL;
    gchar   *pname, *pdate;
    gchar   *filename = "log.gpage";
    gchar   *temp_line;
    
    if ( (fp = get_file_pointer(filename, "r")) == NULL)
	return (1);

    fpt = get_temp_file();
    if ( fpt == NULL )
    {
	g_free(line);
	fclose(fp);
	return (1);
    }	
    line = g_malloc(MAX_LENGTH);

    while( fgets(line, MAX_LENGTH, fp) )
    {
	if ( what == DATE )
	{
	    pdate = get_date(line);
	    if (strcmp(pdate, name) )
		fputs(line, fpt);
	    g_free(pdate);
	}
	else			/* delete by name */
	{
	    pname = get_name(line);
	    if (strcmp(pname, name) )
		fputs(line, fpt);
	    g_free(pname);
	}
    }
    g_free(line);
    fclose(fp);
    rewind(fpt);

    temp_line = g_malloc(MAX_LENGTH);

    if ( (fp = get_file_pointer(filename, "w")) == NULL)
    {
	fclose(fpt);
	g_free(temp_line);
	return(1);
    }
    while(fgets(temp_line, MAX_LENGTH, fpt))
	fputs(temp_line, fp);

    fclose(fp);
    fclose(fpt);

    g_free(temp_line);

    return (return_status);
}

void dump_debug_data(gchar *data)
{
#ifdef FULL_DEBUG

    FILE     *fp;
    gchar    *filename = "client_dump";

    if ( (fp = get_file_pointer(filename, "a")) == NULL)
	return;
    fputs(data, fp);
    fclose(fp);

#endif
}

static void parse_log_line(gchar *line, gchar *status, gchar *via, gchar *date, gchar *message, gchar *name, gchar *code)
{
    int i = 0;
    int x = 0;

    while ( line[i++] != ',' );

    x = 0;
    while ( *(line +i) != ',' )
	*(code + x++) = *(line +i++);

    *(code + x) = '\0';

				/* status */
    if ( code[0] == '0' )
	strcpy(status, "sent");
    else
	strcpy(status, "failed");
				/* via */

    if ( code[1] == '1' )
	strcpy(via, "email");
    else if ( code[1] == '2' )
	strcpy(via, "snpp");
    else if ( code[1] == '3' )
	strcpy(via, "wctp");

				/* date */
    x = 0;
    while ( *(line + i++) != ',');
    i++;
    while ( *(line + i++) != ' ');
    
    while ( (*(date + x++) = *(line + i++)) != ',' );
    *(date + --x) = '\0';
    i++;			/* skip the space */

    x = 0;
    while ( *(line +i) != ',' )
	*(name + x++) = *(line +i++);

    *(name + x) = '\0';
    i++;		
    i++;
    
				/* message */
    x = 0;
    while ( *(line +i) != '\n' )
	*(message + x++) = *(line +i++);

    *(message + x) = '\0';

}





