/*
 * g-page (C) 2001 Hiro Protagonist
 *
 * This software is released under the GNU General Public License.
 * Please read the included LICENSE file for more information.
 * This software comes with no warranty of any kind, use at you own risk!
 */   

#include "my_header.h"
#include "data_structures.h"
#include "function_def.h"

#define MAXFD  64


static void print_command_info();
static int get_mode(int fd);
static banned_ips  *load_banned_ip_list(char *config_file);
static int check_for_banned_ip(char *client_ip, banned_ips *head_ip_list, int mode);
static void report_permission(int fd, int banned);


int   audit_pages;
int   response_poll;
int   debug_flag;
int   daemon_flag;
int   dial_script;
char  *internet_script;

pthread_mutex_t update_client_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t response_poll_mutex = PTHREAD_MUTEX_INITIALIZER;

int main( int argc, char **argv)
{
    enum {
	BLOCK = 1,
	ACCEPT = 2,
	NORMAL = 3
    };

  extern int            debug_flag;
  extern int            daemon_flag;
  extern int            audit_pages;
  extern int            dial_script;
  extern char           *internet_script;

  int                   listen_fd, connected_fd;
  int                   options;
  int                   enable_scheduled_pages = 1;
  int                   i;
  int                   port;
  int                   secure_flag;
  int                   cladr_len; 
  int                   banned;
  int                   optval;
  char                  *resource; 
  char                  *version = "g-page paging server version 2.0";
  char                  *config_file = NULL;
  char                  *client_ip;
  mode_t                mode;
  pid_t                 pid, sch_pid;
  pid_t                 child_pid;
  struct  sockaddr_in   server_adr,client_adr;
  struct  stat          stat_buf;
  banned_ips            *ip_list, *head_ip_list; /* This is a simple linked list of strings */

  ip_list = head_ip_list = NULL;

  debug_flag = 0;
  port = 0;
  banned = 0;			/* If this is one then we are running in secure mode */
  secure_flag = 0;		/* We don't run in secure mode unless told to */
  daemon_flag = 1;		/* We run as a daemon by default */
  dial_script = 0;		/* By default we don't try to manually connect to the Internet */
  audit_pages = 0;
  optval = 0;

  child_pid = pid = sch_pid = 0;

  init_gserver();

  while ( (options = getopt(argc, argv, "adtvsp:f:g:c:")) != -1 )
  {
      switch (options)
      {
      case 'd':
	  debug_flag = 1;
	  info_msg("Debug options set to on");
	  break;

      case 't':
	  daemon_flag = 0;
	  info_msg("Running in terminal mode");
	  break;

      case 'v':
	  daemon_flag = 0;
	  info_msg("%s", version);
	  exit(EXIT_SUCCESS);
	  break;

      case 's':
	  enable_scheduled_pages = 0;
	  info_msg("Scheduled paging disabled");
	  break;

      case 'p':			/* This will override all other port numbers */
	  port = atoi(optarg);
	  info_msg("Using port number: %d\n", port);
	  break;

      case 'f':			/* This is the full path and file name of the file for blocked ip numbers */
	  config_file = get_memory( (strlen(optarg) + 1), "server_main - config_main");
	  strcpy(config_file, optarg);
	  secure_flag = BLOCK;
	  break;

      case 'g':			/* This is the full path and file name of the file for good/accepted ip numbers */
	  config_file = get_memory( (strlen(optarg) + 1), "server_main - config_main");
	  strcpy(config_file, optarg);
	  secure_flag = ACCEPT;
	  break;

      case 'a':
	  audit_pages = 1;
	  break;

      case 'c':
	  dial_script = 1;
	  internet_script = get_memory( (strlen(optarg) + 1), "server_main - config_main");
	  strcpy(internet_script, optarg);
	  break;

      case '?':
	  print_command_info();
	  exit(0);
	  break;
      }
  }

/*
 *   Here is the daemon stuff
 */

  if (daemon_flag)
  {
      openlog("gpage", LOG_PID, LOG_USER);

      if ( ( child_pid = fork() ) == 0 ) /* We are the child  */
      {
	  if ( ( setsid() ) == -1 ) /* Become the session leader */
	  {
	      perror("Error Establishing new session... exiting");
	      exit(EXIT_FAILURE);
	  }

	  Signal(SIGHUP, SIG_IGN);

	  if ( (pid = fork()) != 0 ) /* The first child proc terminates */
	      exit(EXIT_SUCCESS);

	  if ( (chdir("/")) == -1 )
	  {
	      perror("Error changing working directory to root");
	      exit(EXIT_FAILURE);
	  }
	  umask(0);

	  for ( i = 0; i < MAXFD; i++ )
	      close(i);

	  info_msg("gserver daemon started");
      }
      else
	  exit(EXIT_SUCCESS);	/* Parent must exit */
  }

/*
 *   If we are run with the -f option we will load a linked list of all the 
 *   blocked ip numbers and keep them in memory. If we are run with the -g option
 *   we load a list of accepted numbers
 */

  if (secure_flag == BLOCK)
  {
      ip_list = load_banned_ip_list(config_file);
      if ( ip_list == NULL )
	  error_quit("Error reading from specified file: %s\n", config_file);
  }
  else if (secure_flag == ACCEPT)
  {
      ip_list = load_banned_ip_list(config_file);
      if ( ip_list == NULL )
	  error_quit("Error reading from specified file: %s\n", config_file);
  }
  else
      secure_flag = NORMAL;


/*
 *    We fork off a new process to handle schedule pages here unless
 *    we were started with the -s option
 */

  if (  enable_scheduled_pages )
      if ( ( sch_pid = fork() ) == 0 )
      {
	  handle_scheduled_pages();
	  exit(EXIT_SUCCESS);
      }

/*
 *  Okay, it's time to create a listening socket for g-page clients
 *  to connect to.
 */
  
  if ( (listen_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      error_sys("Error creating socket - server_main.c ");
    
  memset(&server_adr, '0', sizeof(server_adr));
  server_adr.sin_family = AF_INET;
  server_adr.sin_addr.s_addr = htonl(INADDR_ANY);

  if ( port == 0 )		/* The user did not set the port from the command line */
  {
      memset(&stat_buf, 0, sizeof(stat_buf));
      if ( stat("/etc/gpage.conf", &stat_buf) == -1 )
      {
	  info_msg("gpage.conf does not exist...using default port %d", DEFAULT_PORT);
	  server_adr.sin_port = htons(DEFAULT_PORT); 
      }
      else
      {
	  resource = get_memory(MAX_LENGTH, "server_main - resource");
	  if ( (get_resource_data("/etc/gpage.conf", "port", resource)) != 1 )
	  {
	      info_msg("Error reading gpage.conf.. trying default port %d", DEFAULT_PORT);
	      server_adr.sin_port = htons(DEFAULT_PORT);
	  }
	  else
	      server_adr.sin_port = htons(atoi(resource)); /* The port number is defined here */
	  free(resource);
      }
  }
  else
	  server_adr.sin_port = htons(port); /* Use the port set as a command line option */

  optval = 1;			/* This turns on the SO_REUSEADDR option - a zero would turn it off */
  if ( setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) == -1 )
      info_msg("Error setting socket options - attempting to continue anyways"); /* This is not a showstopper if this fails */

  if ( (bind(listen_fd, (struct sockaddr *) &server_adr, sizeof(server_adr))) < 0 )
    error_sys("Error binding address - server_main.c ");
    
  if ( (listen(listen_fd, 5)) < 0 )
    error_sys("Error in listen - server_main.c ");
  
  Signal(SIGCHLD, sig_chld);

/*
 *  Here is the main loop that runs forever waiting for client connections
 *
 *  After we connect we check to see if the client ip is on the blocked or accepted list, 
 *  if the ip is okay we get the "mode" from the client and call the appropriate
 *  funtion.
 */

  for (;;)
  {
      if ( (connected_fd = accept(listen_fd, (struct sockaddr *) NULL, NULL)) < 0 )
      {
	   if ( errno == EINTR )
		continue;
	   else
		error_sys("Error in accept - server_main.c ");
      }
      
      if ( (child_pid = fork() ) == 0) /* This is the child process */
      {
	  close(listen_fd);

	  cladr_len = sizeof(client_adr);
	  if ( (getpeername(connected_fd, (struct sockaddr *)&client_adr, &cladr_len) == -1) )
	      error_ret("Error looking up client ip address - server_main.c");

	  client_ip = get_memory(IP_LENGTH, "server_main - client_ip");
	  strcpy(client_ip, inet_ntoa(client_adr.sin_addr));
	  debug_log("Connect with remote host...%s", client_ip);

	  head_ip_list = ip_list;	  

	  /* We get a one back if the client ip address is on the blocked list */
	  banned = 0;

	  if (secure_flag == BLOCK)
	      banned = check_for_banned_ip(client_ip, head_ip_list, BLOCK);
	  else if (secure_flag == ACCEPT)
	      banned = check_for_banned_ip(client_ip, head_ip_list, ACCEPT);

	  report_permission(connected_fd, banned);

	  if (banned)
	  {
	      close(connected_fd);	   
	      closelog();
	      exit(0);
	  }
	  mode = get_mode(connected_fd);
	  switch (mode)
	  {
	  case STANDARD:
	      gui_page(connected_fd, client_ip);
	      break;
	  case DIRECT:
	      auto_page(connected_fd, client_ip);
	      break;
	  case STATUS:		/* Get the message status of a scheduled page */
	      get_msg_status(connected_fd);
	      break;
	  case DELETE:		/* Delete a scheduled page  */
	      delete_msg(connected_fd);
	      break;
	  case TELNET:
	      telnet_page(connected_fd, client_ip);
	      break;
	  default:
	    debug_log("Error: invalid mode: %d", mode);
	    break;
	  }

	  closelog();
	  free(client_ip);
	  //verify_exit();
	  exit(0);		/* This is where the child exits after sending the pages*/
      }
      else
	  close(connected_fd);		/* This is the parent process */
  }
  
  exit(EXIT_FAILURE);		/* If we get here something evil has happened */
}

/*
 *   The mode is the first thing sent to the server from the client. It is either
 *   a number or a word. 
 */

static int get_mode(int fd)
{
    int       num_read = 0;
    int       mode = 0;
    char      *buffer = "";

    buffer = get_memory(MAX_LENGTH, "server_main - buffer");

    num_read = read_a_line(fd, buffer, MAX_LENGTH, MAX_TIMEOUT);
    
    /*
     *   This is a temporary kludge. If we get an invalid mode we assume its a zero
     */

    if ( num_read == 0 )
	strcpy(buffer, "0\n");

    write_to_status_line(fd, buffer, num_read, IGNORE);
    
    if ( !strcmp(buffer, "page\n\r") || !strcmp(buffer, "page\r\n") || !strcmp(buffer, "page\n") )
    {
	free(buffer);
	return(TELNET);
    }

    mode = (atoi(buffer));
    free(buffer);
    return (mode);
}

void report_permission(int fd, int banned)
{
    if (banned)
	write_to_status_line(fd, "denied\n", strlen("denied\n"), SPECIAL);
    else
	write_to_status_line(fd, "okay\n", strlen("okay\n"), SPECIAL);
}

/*
 *   We create the ~/.gpage directory if it doesn't exist and we create the SCH_ID.dat
 *   file if it doens't exist. We exit on failure.
 */

void init_gserver()
{
    extern int            response_poll;
    FILE                  *fp;
    char                  *home_dir = ""; /* The home directory for g-page: ~/.gpage */
    struct  stat          stat_buf;
    mode_t                smode;
    int                   path_length = 0;

    response_poll = 0;

    path_length = strlen(getenv("HOME")) + strlen("/.gpage") + 10;
    
    home_dir = get_memory(MAX_LENGTH, "server_main - home_dir");

    strcpy(home_dir, getenv("HOME"));
    strcat(home_dir, "/.gpage");
    memset(&stat_buf, 0, sizeof(stat_buf));
    stat(home_dir, &stat_buf);
    smode = stat_buf.st_mode;
    if (! S_ISDIR(smode))
    {
	if ( (mkdir(home_dir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP
		    | S_IXGRP | S_IXOTH)) == -1)
	{
	    perror("Error creating .gpage directory: ");
	    free(home_dir);
	    exit(EXIT_FAILURE);
	}
	else
	{
	    strcat(home_dir, "/SCH_ID.dat");
	    if ( ( fp = fopen(home_dir, "w+")) == NULL)
		perror("Error creating SCH_ID.dat ");
	    fputs("-1\n", fp);
	    fputs("SCHEDULE,2\n", fp);
	    fclose(fp);
	}
    }
    else
    {
	strcat(home_dir, "/SCH_ID.dat");
	memset(&stat_buf, 0, sizeof(stat_buf));
	stat(home_dir, &stat_buf);
	smode = stat_buf.st_mode;
	if (! S_ISREG(smode) )
	{
	    if ( ( fp = fopen(home_dir, "w+")) == NULL)
	    {
		perror("Error creating SCH_ID.dat ");
		exit(EXIT_FAILURE);
	    }
	    fputs("-1\n", fp);
	    fputs("SCHEDULE,2\n", fp);
	    fclose(fp);
	}
    }

    free(home_dir);
}

void sig_chld(int signo)
{
  pid_t     pid;
  int       stat;
  char      temp[128];

  memset(temp, '\0', sizeof(temp));

  while ( (pid = waitpid(-1, &stat, WNOHANG) ) > 0)
  {
      sprintf(temp, "Child Process %u has terminated", (int )pid);
      //print_debug_data(temp);
  }

  return;
}

static void print_command_info()
{
    fprintf(stderr, "G-page server command line options:\n\n");
    fprintf(stderr, "-d Run the server in debug mode. This will log debug data into your ~/.gpage directory\n");
    fprintf(stderr, "-t Run the server in a terminal. Use this switch to stop the server from running as a daemon\n");
    fprintf(stderr, "-v Print the version number and exit\n");
    fprintf(stderr, "-s Disable the scheduled paging server. Use the switch if you don't need to send scheduled pages\n");
    fprintf(stderr, "-p [port #] Specify a port number for the server to listen on, overriding the one specified in the config file\n");
    fprintf(stderr, "-f [filename] Specifiy a filename for the list of banned ip's to be stored in. This must be the full path\n");
    fprintf(stderr, "-g [filename] Specifiy a filename for the list of accepted ip's to be stored in. This must be the full path\n");
    fprintf(stderr, "-a Enable client auditing, all pages are logged with the client ip address\n");
    fprintf(stderr, "-? Print this help message and exit\n");
}

/*
 *  We look read the specified file and load the ip numbers from that file
 *  into a linked list. We return NULL on error.
 */

static banned_ips  *load_banned_ip_list(char *config_file)
{
    FILE                  *fp;
    banned_ips            *list, *head, *temp;
    struct  stat          stat_buf;
    mode_t                smode;
    
    if ( (list = malloc(sizeof(*list))) == NULL )
	exit(EXIT_FAILURE);
    head = list;
    list->next = NULL;

    stat(config_file, &stat_buf);
    smode = stat_buf.st_mode;
    if (S_ISREG(smode))
    {
	if ( (fp = fopen(config_file, "r")) == NULL)
	    error_quit("Error opening config file: %s", config_file);
	list->bad_ip = get_memory(IP_LENGTH, "server_main - list->bad_ip");
	while ( fgets(list->bad_ip, IP_LENGTH, fp) )
	{
	    if ( strlen(list->bad_ip) == 1 ) /* We are at the end of the file, or it's corrupt */
		continue;
	    list->bad_ip[strlen(list->bad_ip) -1] = '\0'; /* Kill the newline */
	    if ( (temp = malloc(sizeof(*temp))) == NULL )
		exit(EXIT_FAILURE);
	    temp->bad_ip = get_memory(IP_LENGTH, "server_main - temp->bad_ip");
	    temp->next = NULL;
	    list->next = temp;
	    list = temp;
	}
	fclose(fp);
    }
    else			/* The file is not a "regular file" this is an error */
    {
	free(list);
	return(NULL);
    }
    return (head);

}

/*
 *   We go through the linked list looking for a match with the ip number
 *   loaded into client_ip. We need to see if we are looking for a blocked #ip
 *   or an accepted #ip, based upon the mode. We return a one if the #ip is banned,
 *   we return a zero if it's good.
 *
 *   FUTURE - set this up to work with wildcards
 */

static int check_for_banned_ip(char *client_ip, banned_ips *head_ip_list, int mode)
{
#define  MAX_DOMAIN_NAME 512

    enum {
	BLOCK = 1,
	ACCEPT = 2,
    };

    struct in_addr   server_adr;
    struct hostent   *host_data;
    char             domain[MAX_DOMAIN_NAME] =  "";
    
/*
 *   If inet_aton or gethostbyaddr have errors I'm going to ignore them because we 
 *   can still try to compare against the client ip address. If they do have an error
 *   we will be comparing against an empty string so it should be okay
 */

    if ( !(inet_aton(client_ip, &server_adr)) )
	info_msg("Error converting client address to network address: %s\n", client_ip);
    else
    {
	if ( !(host_data = gethostbyaddr( (char *) &server_adr, sizeof(server_adr), AF_INET)) )
	    info_msg("Error looking up client address - gethostbyaddr: %s\n", client_ip);
	else
	{
	    if ( strlen(host_data->h_name) < ( MAX_DOMAIN_NAME - 1 ) )
		 strcpy(domain, host_data->h_name);
	    else
		info_msg("Error: doamin name is larger than %d\n", MAX_DOMAIN_NAME);
	}
    }

    if ( mode == BLOCK )
    {
	while (head_ip_list)
	{
	    /* info_msg("READ IP: %s\n", head_ip_list->bad_ip); */
	    if (  (!strcmp(client_ip, head_ip_list->bad_ip)) || (!strcmp(domain, head_ip_list->bad_ip)) )
		//if (!strcmp(client_ip, head_ip_list->bad_ip) )
		return (1);
	    head_ip_list = head_ip_list->next;
	}
	return(0);
    }
    else if ( mode == ACCEPT )
    {
	while (head_ip_list)
	{
	    /* info_msg("READ IP: %s\n", head_ip_list->bad_ip); */
	    if (!strcmp(client_ip, head_ip_list->bad_ip) )
		return (0);
	    head_ip_list = head_ip_list->next;
	}
	return(1);
    }
    return (1);			/* Should never get here */
}

void verify_exit()
{
 extern pthread_mutex_t response_poll_mutex;
 extern int response_poll;
 int    not_safe, mutex_status;
 int    exit_count;

 not_safe = 1;
 exit_count = 0;

 while (not_safe)
 {
     if ( (mutex_status = pthread_mutex_lock(&response_poll_mutex)) != 0 )
	 error_quit("Error locking mutex - possible data corruption");

     if (response_poll == 0)
	 not_safe = 0;

     if ( (mutex_status = pthread_mutex_unlock(&response_poll_mutex)) != 0 )
	 error_quit("Error unlocking mutex - possible data corruption");

     if (not_safe)
	 sleep(15);

     exit_count++;
     if (exit_count == 150)	/* Just in case something evil happens - this will stop us from hanging forever */
	 break;
 }

}
