/*
 * g-page (C) 2001 Hiro Protagonist
 *
 * This software is released under the GNU General Public License.
 * Please read the included LICENSE file for more information.
 * This software comes with no warranty of any kind, use at you own risk!
 */   

#include "my_header.h"

#define MAXFD  64
#define PORT_NUMBER 1301

static void print_command_info();
static banned_ips  *load_banned_ip_list(char *config_file);
static int check_for_banned_ip(char *client_ip, banned_ips *head_ip_list, int mode);


pthread_mutex_t update_client_mutex = PTHREAD_MUTEX_INITIALIZER;

int main( int argc, char **argv)
{
    enum {
	BLOCK = 1,
	ACCEPT = 2,
	NORMAL = 3
    };

  extern int            debug_flag;
  extern int            daemon_flag;

  int                   listen_fd, connected_fd;
  int                   options;
  int                   i;
  int                   port;
  int                   secure_flag;
  int                   cladr_len; 
  int                   banned;
  char                  *version = "g-page database server version 2.0";
  char                  *config_file = "";
  char                  *client_ip;
  pid_t                 pid, sch_pid;
  pid_t                 child_pid;
  struct  sockaddr_in   server_adr,client_adr;
  banned_ips            *ip_list, *head_ip_list; /* This is a simple linked list of strings */

  ip_list = NULL;
  head_ip_list = NULL;
  debug_flag = 0;
  port = 0;
  banned = 0;			/* If this is one then we are running in secure mode */
  secure_flag = 0;		/* We don't run in secure mode unless told to */
  daemon_flag = 1;		/* We run as a daemon by default */

  child_pid = pid = sch_pid = 0;

  while ( (options = getopt(argc, argv, "dtvp:f:g:")) != -1 )
  {
      switch (options)
      {
      case 'd':
	  debug_flag = 1;
	  info_msg("Debug options set to on");
	  break;

      case 't':
	  daemon_flag = 0;
	  info_msg("Running in terminal mode");
	  break;

      case 'v':
	  daemon_flag = 0;
	  info_msg("%s", version);
	  exit(EXIT_SUCCESS);
	  break;

      case 'p':			/* This will override all other port numbers */
	  port = atoi(optarg);
	  info_msg("Using port number: %d\n", port);
	  break;

      case 'f':			/* This is the full path and file name of the file for blocked ip numbers */
	  config_file = get_memory( (strlen(optarg) + 1), "server_main - config_main");
	  strcpy(config_file, optarg);
	  secure_flag = BLOCK;
	  break;

      case 'g':			/* This is the full path and file name of the file for good/accepted ip numbers */
	  config_file = get_memory( (strlen(optarg) + 1), "server_main - config_main");
	  strcpy(config_file, optarg);
	  secure_flag = ACCEPT;
	  break;

      case '?':
	  print_command_info();
	  exit(0);
	  break;
      }
  }

/*
 *   Here is the daemon stuff
 */

  if (daemon_flag)
  {
      openlog("gpage", LOG_PID, LOG_USER);

      if ( ( child_pid = fork() ) == 0 ) /* We are the child  */
      {
	  if ( ( setsid() ) == -1 ) /* Become the session leader */
	  {
	      perror("Error Establishing new session... exiting");
	      exit(EXIT_FAILURE);
	  }

	  Signal(SIGHUP, SIG_IGN);

	  if ( (pid = fork()) != 0 ) /* The first child proc terminates */
	      exit(EXIT_SUCCESS);

	  if ( (chdir("/")) == -1 )
	  {
	      perror("Error changing working directory to root");
	      exit(EXIT_FAILURE);
	  }
	  umask(0);

	  for ( i = 0; i < MAXFD; i++ )
	      close(i);

	  info_msg("db_server daemon started");
      }
      else
	  exit(EXIT_SUCCESS);	/* Parent must exit */
  }

/*
 *   If we are run with the -f option we will load a linked list of all the 
 *   blocked ip numbers and keep them in memory. If we are run with the -g option
 *   we load a list of accepted numbers
 */

  if (secure_flag == BLOCK)
  {
      ip_list = load_banned_ip_list(config_file);
      if ( ip_list == NULL )
	  error_quit("Error reading from specified file: %s\n", config_file);
  }
  else if (secure_flag == ACCEPT)
  {
      ip_list = load_banned_ip_list(config_file);
      if ( ip_list == NULL )
	  error_quit("Error reading from specified file: %s\n", config_file);
  }
  else
      secure_flag = NORMAL;


/*
 *  Okay, it's time to create a listening socket
 */
  
  if ( (listen_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      error_sys("Error creating socket - server_main.c ");
    
  memset(&server_adr, '0', sizeof(server_adr));
  server_adr.sin_family = AF_INET;
  server_adr.sin_addr.s_addr = htonl(INADDR_ANY);

  if ( port == 0 )		/* The user did not set the port from the command line */
      server_adr.sin_port = htons(PORT_NUMBER);
  else
	  server_adr.sin_port = htons(port); /* Use the port set as a command line option */

  if ( (bind(listen_fd, (struct sockaddr *) &server_adr, sizeof(server_adr))) < 0 )
    error_sys("Error binding address - server_main.c ");
    
  if ( (listen(listen_fd, 5)) < 0 )
    error_sys("Error in listen - server_main.c ");
  
  Signal(SIGCHLD, sig_chld);

/*
 *  After we connect we check to see if the client ip is on the blocked list, 
 *  if the ip is okay we get the "mode" from the client and call the appropriate
 *  funtion.
 */

  for (;;)
  {
      if ( (connected_fd = accept(listen_fd, (struct sockaddr *) NULL, NULL)) < 0 )
      {
	   if ( errno == EINTR )
		continue;
	   else
		error_sys("Error in accept - server_main.c ");
      }
      
      if ( (child_pid = fork() ) == 0) /* This is the child process */
      {
	  close(listen_fd);

	  cladr_len = sizeof(client_adr);
	  if ( (getpeername(connected_fd, (struct sockaddr *)&client_adr, &cladr_len) == -1) )
	      error_ret("Error looking up client ip address - server_main.c");

	  client_ip = get_memory(IP_LENGTH, "server_main - client_ip");
	  strcpy(client_ip, inet_ntoa(client_adr.sin_addr));
	  debug_log("Connect with remote host...%s", client_ip);


	  head_ip_list = ip_list;	  

	  /* We get a one back if the client ip address is on the blocked list */
	  banned = 0;

	  if (secure_flag == BLOCK)
	      banned = check_for_banned_ip(client_ip, head_ip_list, BLOCK);
	  else if (secure_flag == ACCEPT)
	      banned = check_for_banned_ip(client_ip, head_ip_list, ACCEPT);

	  if (banned)
	  {
	      write_num(connected_fd, "denied\n", strlen("denied\n"));
	      close(connected_fd);	   
	      closelog();
	      free(client_ip);
	      exit(0);
	  }
	  else
	      write_num(connected_fd, "okay\n", strlen("okay\n"));

	  get_data(connected_fd, client_ip);

	  closelog();
	  free(client_ip);
	  exit(0);	
      }
      else
	  close(connected_fd);		/* This is the parent process */
  }
  
  exit(EXIT_FAILURE);		/* If we get here something evil has happened */
}


void sig_chld(int signo)
{
  pid_t     pid;
  int       stat;
  char      temp[128];

  memset(temp, '\0', sizeof(temp));

  while ( (pid = waitpid(-1, &stat, WNOHANG) ) > 0)
      sprintf(temp, "Child Process %u has terminated", (int )pid);

  return;
}

static void print_command_info()
{
    fprintf(stderr, "G-page db_server command line options:\n\n");
    fprintf(stderr, "-t Run the db_server in a terminal. Use this switch to stop the db_server from running as a daemon\n");
    fprintf(stderr, "-v Print the version number and exit\n");
    fprintf(stderr, "-p [port #] Specify a port number for the db_server to listen on, overriding 1301\n");
    fprintf(stderr, "-f [filename] Specifiy a filename for the list of banned ip's to be stored in. This must be the full path\n");
    fprintf(stderr, "-g [filename] Specifiy a filename for the list of accepted ip's to be stored in. This must be the full path\n");
    fprintf(stderr, "-? Print this help message and exit\n");
}

/*
 *  We look read the specified file and load the ip numbers from that file
 *  into a linked list. We return NULL on error.
 */

static banned_ips  *load_banned_ip_list(char *config_file)
{
    FILE                  *fp;
    banned_ips            *list, *head, *temp;
    struct  stat          stat_buf;
    mode_t                smode;
    
    if ( (list = malloc(sizeof(*list))) == NULL )
	exit(EXIT_FAILURE);
    head = list;
    list->next = NULL;

    stat(config_file, &stat_buf);
    smode = stat_buf.st_mode;
    if (S_ISREG(smode))
    {
	if ( (fp = fopen(config_file, "r")) == NULL)
	    error_quit("Error opening config file: %s", config_file);
	list->bad_ip = get_memory(IP_LENGTH, "server_main - list->bad_ip");
	while ( fgets(list->bad_ip, IP_LENGTH, fp) )
	{
	    if ( strlen(list->bad_ip) == 1 ) /* We are at the end of the file, or it's corrupt */
		continue;
	    list->bad_ip[strlen(list->bad_ip) -1] = '\0'; /* Kill the newline */
	    if ( (temp = malloc(sizeof(*temp))) == NULL )
		exit(EXIT_FAILURE);
	    temp->bad_ip = get_memory(IP_LENGTH, "server - temp->bad_ip");
	    temp->next = NULL;
	    list->next = temp;
	    list = temp;
	}
	fclose(fp);
    }
    else			/* The file is not a "regular file" this is an error */
    {
	free(list);
	return(NULL);
    }
    return (head);

}

/*
 *   We go through the linked list looking for a match with the ip number
 *   loaded into client_ip. We need to see if we are looking for a blocked #ip
 *   or an accepted #ip, based upon the mode. We return a one if the #ip is banned,
 *   we return a zero if it's good.
 *
 *   FUTURE - set this up to work with wildcards or perhaps domain names
 */

static int check_for_banned_ip(char *client_ip, banned_ips *head_ip_list, int mode)
{
#define  MAX_DOMAIN_NAME 512

    enum {
	BLOCK = 1,
	ACCEPT = 2,
    };

    struct in_addr   server_adr;
    struct hostent   *host_data;
    char             domain[MAX_DOMAIN_NAME] =  "";
    
/*
 *   If inet_aton or gethostbyaddr have errors I'm going to ignore them because we 
 *   can still try to compare against the client ip address. If they do have an error
 *   we will be comparing against an empty string so it should be okay
 */

    if ( !(inet_aton(client_ip, &server_adr)) )
	info_msg("Error converting client address to network address: %s\n", client_ip);
    else
    {
	if ( !(host_data = gethostbyaddr( (char *) &server_adr, sizeof(server_adr), AF_INET)) )
	    info_msg("Error looking up client address - gethostbyaddr: %s\n", client_ip);
	else
	{
	    if ( strlen(host_data->h_name) < (MAX_DOMAIN_NAME -1) )
		strcpy(domain, host_data->h_name);
	    else
		info_msg("Error domain name (%s) exceeds maximum length allowed (%d)\n", host_data->h_name, MAX_DOMAIN_NAME);
	}
    }

    if ( mode == BLOCK )
    {
	while (head_ip_list)
	{
	    /* info_msg("READ IP: %s\n", head_ip_list->bad_ip); */
	    if (  (!strcmp(client_ip, head_ip_list->bad_ip)) || (!strcmp(domain, head_ip_list->bad_ip)) )
		return (1);
	    head_ip_list = head_ip_list->next;
	}
	return(0);
    }
    else if ( mode == ACCEPT )
    {
	while (head_ip_list)
	{
	    /* info_msg("READ IP: %s\n", head_ip_list->bad_ip); */
	    if (!strcmp(client_ip, head_ip_list->bad_ip) )
		return (0);
	    head_ip_list = head_ip_list->next;
	}
	return(1);
    }
    return (1);			/* Should never get here */
}


char *get_memory(int size, char *function)
{
    char *str;
    static int total = 0;
 
    total += size;
    //fprintf(stderr, "Here's the memory toal: %d\n", total);
    if ( (str = malloc(size)) == NULL)
        error_quit("Error allcoating memory %s ", function);
 
    memset(str, '\0', size);
    /* fprintf(stderr, "Allocating: %s - %d\n", function, size); */
    return str;
}        

Sigfunc *Signal(int signo, Sigfunc *func)
{
 
    struct sigaction  act, oact;
 
    act.sa_handler = func;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if (  signo == SIGALRM )
    {
#ifdef SA_INTERRUPT
        act.sa_flags |= SA_INTERRUPT;
#endif
    }
    else
    {
#ifdef SA_RESTART
        act.sa_flags |= SA_RESTART;
#endif
    }
    if ( sigaction(signo, &act, &oact) < 0 )
        return(SIG_ERR);
    return (oact.sa_handler);
 
}         
